#include "defines.h"
#include "kozos.h"
#include "intr.h"
#include "interrupt.h"
#include "syscall.h"
#include "lib.h"

#define THREAD_NUM 6
#define THREAD_NAME_SIZE 15

/* XbhEReLXg */
typedef struct _kz_context {
  uint32 sp; /* X^bNE|C^ */
} kz_context;

/* Xbh\ */
typedef struct _kz_thread {
  struct _kz_thread *next;
  char name[THREAD_NAME_SIZE + 1]; /* Xbh */
  kz_func_t func; /* Xbhmain()֐ */
  char *stack;    /* X^bN */

  struct { /* VXeER[pobt@ */
    kz_syscall_type_t type;
    kz_syscall_param_t *param;
  } syscall;

  kz_context context; /* ReLXg */
} kz_thread;

/* Xbh̃fB[EL[ */
static struct {
  kz_thread *head;
  kz_thread *tail;
} readyque;

static kz_thread *current; /* JgEXbh */
static kz_thread threads[THREAD_NUM]; /* Xbh\ */
static kz_handler_t handlers[SOFTVEC_TYPE_NUM]; /* ݃nh */

void dispatch(kz_context *context);

/* JgEXbhfB[EL[甲o */
static int getcurrent(void)
{
  if (current == NULL) {
    return -1;
  }

  /* JgEXbh͕K擪ɂ͂Ȃ̂ŁC擪甲o */
  readyque.head = current->next;
  if (readyque.head == NULL) {
    readyque.tail = NULL;
  }
  current->next = NULL;

  return 0;
}

/* JgEXbhfB[EL[Ɍq */
static int putcurrent(void)
{
  if (current == NULL) {
    return -1;
  }

  /* fB[EL[̖ɐڑ */
  if (readyque.tail) {
    readyque.tail->next = current;
  } else {
    readyque.head = current;
  }
  readyque.tail = current;

  return 0;
}

static void thread_end(void)
{
  kz_exit();
}

/* Xbh̃X^[gEAbv */
static void thread_init(kz_thread *thp, int argc, char *argv[])
{
  thp->func(argc, argv); /* Xbhmain()֐Ăяo */
  thread_end();
}

/* VXeER[̏(kz_run():Xbh̋N) */
static kz_thread_id_t thread_run(kz_func_t func, char *name,
				 int stacksize, int argc, char *argv[])
{
  int i;
  kz_thread *thp;
  uint32 *sp;
  extern char userstack; /* JEXNvgŒ`X^bN̈ */
  static char *thread_stack = &userstack;

  /* 󂢂ĂXbh\̂ */
  for (i = 0; i < THREAD_NUM; i++) {
    thp = &threads[i];
    if (!thp->func) /*  */
      break;
  }
  if (i == THREAD_NUM) /* Ȃ */
    return -1;

  memset(thp, 0, sizeof(*thp));

  /* Xbh\̂̐ݒ */
  strcpy(thp->name, name);
  thp->next = NULL;
  thp->func = func;

  /* X^bN̈l */
  memset(thread_stack, 0, stacksize);
  thread_stack += stacksize;

  thp->stack = thread_stack; /* X^bNݒ */

  /* X^bN̏ */
  sp = (uint32 *)thp->stack;
  *(--sp) = (uint32)thread_end;

  /*
   * vOEJE^ݒ肷D
   */
  *(--sp) = (uint32)thread_init;

  *(--sp) = 0; /* ER6 */
  *(--sp) = 0; /* ER5 */
  *(--sp) = 0; /* ER4 */
  *(--sp) = 0; /* ER3 */

  /* Xbh̃X^[gEAbv(thread_init())ɓn */
  *(--sp) = (uint32)argv; /* ER2 */
  *(--sp) = (uint32)argc; /* ER1 */
  *(--sp) = (uint32)thp;  /* ER0 */

  /* Xbh̃ReLXgݒ */
  thp->context.sp = (uint32)sp;

  /* VXeER[ĂяoXbhfB[EL[ɖ߂ */
  putcurrent();

  /* VK쐬XbhCfB[EL[ɐڑ */
  current = thp;
  putcurrent();

  return (kz_thread_id_t)current;
}

/* VXeER[̏(kz_exit():Xbh̏I) */
static int thread_exit(void)
{
  /*
   * {ȂX^bNčėpł悤ɂׂȗD
   * ̂߁CXbhpɂɐE悤ȂƂ͌łłȂD
   */
  puts(current->name);
  puts(" EXIT.\n");
  memset(current, 0, sizeof(*current));
  return 0;
}

/* ݃nh̓o^ */
static int setintr(softvec_type_t type, kz_handler_t handler)
{
  static void thread_intr(softvec_type_t type, unsigned long sp);

  /*
   * ݂󂯕t邽߂ɁC\tgEGAE݃xN^
   * OS̊ݏ̓ƂȂ֐o^D
   */
  softvec_setintr(type, thread_intr);

  handlers[type] = handler; /* OSĂяo݃nho^ */
  putcurrent();

  return 0;
}

static void call_functions(kz_syscall_type_t type, kz_syscall_param_t *p)
{
  /* VXeR[̎scurrent̂Œ */
  switch (type) {
  case KZ_SYSCALL_TYPE_RUN: /* kz_run() */
    p->un.run.ret = thread_run(p->un.run.func, p->un.run.name,
			       p->un.run.stacksize,
			       p->un.run.argc, p->un.run.argv);
    break;
  case KZ_SYSCALL_TYPE_EXIT: /* kz_exit() */
    /* Xbh\̂̂ŁC߂lł͂Ȃ */
    thread_exit();
    break;
  default:
    break;
  }
}

/* VXeER[̏ */
static void syscall_proc(kz_syscall_type_t type, kz_syscall_param_t *p)
{
  /*
   * VXeER[ĂяoXbhfB[EL[
   * OԂŏ֐ĂяoD̂߃VXeER[
   * ĂяoXbĥ܂ܓpꍇɂ́C
   * ֐̓ putcurrent() sKvD
   */
  getcurrent();
  call_functions(type, p);
}

/* Xbh̃XPW[O */
static void schedule(void)
{
  if (!readyque.head) /* Ȃ */
    kz_sysdown();

  current = readyque.head; /* JgEXbhɐݒ肷 */
}

static void syscall_intr(void)
{
  syscall_proc(current->syscall.type, current->syscall.param);
}

static void softerr_intr(void)
{
  puts(current->name);
  puts(" DOWN.\n");
  getcurrent(); /* fB[L[O */
  thread_exit(); /* XbhI */
}

/* ݏ̓֐ */
static void thread_intr(softvec_type_t type, unsigned long sp)
{
  /* JgEXbh̃ReLXgۑ */
  current->context.sp = sp;

  /*
   * ݂Ƃ̏sD
   * SOFTVEC_TYPE_SYSCALL, SOFTVEC_TYPE_SOFTERR ̏ꍇ
   * syscall_intr(), softerr_intr() nhɓo^Ă̂ŁC
   * 炪sD
   */
  if (handlers[type])
    handlers[type]();

  schedule(); /* Xbh̃XPW[O */

  /*
   * Xbh̃fBXpb`
   * (dispatch()֐̖{̂startup.sɂCAZuŋLqĂ)
   */
  dispatch(&current->context);
  /* ɂ͕ԂĂȂ */
}

void kz_start(kz_func_t func, char *name, int stacksize,
	      int argc, char *argv[])
{
  /*
   * ȍ~ŌĂяoXbh֘ÃCu֐̓ current 
   * Ăꍇ̂ŁCcurrent  NULL ɏĂD
   */
  current = NULL;

  readyque.head = readyque.tail = NULL;
  memset(threads,  0, sizeof(threads));
  memset(handlers, 0, sizeof(handlers));

  /* ݃nh̓o^ */
  setintr(SOFTVEC_TYPE_SYSCALL, syscall_intr); /* VXeER[ */
  setintr(SOFTVEC_TYPE_SOFTERR, softerr_intr); /* _Ev */

  /*
   * current ̂߂ɃVXeR[s͂łȂ̂ŁC
   * ڊ֐ĂяoăXbh쐬D
   */
  current = (kz_thread *)thread_run(func, name, stacksize, argc, argv);

  /* ŏ̃XbhN */
  dispatch(&current->context);

  /* ɂ͕ԂĂȂ */
}

void kz_sysdown(void)
{
  puts("system error!\n");
  while (1)
    ;
}

/* VXeER[ĂяopCu֐ */
void kz_syscall(kz_syscall_type_t type, kz_syscall_param_t *param)
{
  current->syscall.type  = type;
  current->syscall.param = param;
  asm volatile ("trapa #0"); /* gbvݔs */
}
