#include "defines.h"
#include "kozos.h"
#include "intr.h"
#include "interrupt.h"
#include "serial.h"
#include "lib.h"
#include "consdrv.h"

#define CONS_BUFFER_SIZE 24

/*
 * ӁI H816rbgCPUȂ̂ŁC32rbg̊|ZłȂD
 * |C^32ƂĈ̂ŁCconsreg[x] ̂悤ȔzANZXɂ
 * 32rbg̊|ZgpD
 * ̂ߔzQƂ\̂́CTCYQ̊KɂȂ悤ɒ邱ƁD
 * (TCYQ̊KȂ΁C|ZłȂVtgZŔzANZX悤
 *  R[hRpCo͂DȂꍇ̓_~[oǉăTCY
 *  邱)
 */
static struct consreg {
  kz_thread_id_t id; /* f[^M̒ʒmXbh */
  int index;         /* pVA̔ԍ */

  char *send_buf;    /* Mobt@ */
  char *recv_buf;    /* Mobt@ */
  int send_len;      /* Mobt@̃f[^TCY */
  int recv_len;      /* Mobt@̃f[^TCY */

  uint32 dummy[3];   /* p_~[o */
} consreg[CONSDRV_DEVICE_NUM];

/*
 * ȉ̂Q̊֐(send_char(), send_string())͊ݏƃXbh
 * Ă΂邪Mobt@𑀍삵Ăēs̂߁CXbhĂяo
 * ꍇ͔r̂ߊ݋֎~ԂŌĂԂƁD
 */

/* Mobt@̐擪P𑗐M */
static void send_char(struct consreg *cons)
{
  int i;
  serial_send_byte(cons->index, cons->send_buf[0]);
  cons->send_len--;
  /* 擪𑗐M̂ŁCPԂ񂸂炷 */
  for (i = 0; i < cons->send_len; i++)
    cons->send_buf[i] = cons->send_buf[i + 1];
}

/* 𑗐Mobt@ɏݑMJn */
static void send_string(struct consreg *cons, char *str, int len)
{
  int i;
  for (i = 0; i < len; i++) { /* 𑗐Mobt@ɃRs[ */
    if (str[i] == '\n') /* \n\r\nɕϊ */
      cons->send_buf[cons->send_len++] = '\r';
    cons->send_buf[cons->send_len++] = str[i];
  }
  /*
   * MݖȂ΁CMJnĂȂ̂őMJnD
   * MݗLȂΑMJnĂCM݂̉
   * Mobt@̃f[^M̂ŁCȂĂ悢D
   */
  if (cons->send_len && !serial_intr_is_send_enable(cons->index)) {
    serial_intr_send_enable(cons->index); /* MݗL */
    send_char(cons); /* MJn */
  }
}

/*
 * ȉ͊݃nhĂ΂銄ݏłC񓯊
 * Ă΂̂ŁCCu֐ȂǂĂяoꍇɂ͒ӂKvD
 * {ƂāCȉ̂ꂩɓĂ͂܂֐ĂяoĂ͂ȂD
 * Eē\łD
 * EXbhĂ΂邱Ƃ͖֐łD
 * EXbhĂ΂邱Ƃ邪C݋֎~ŌĂяoĂD
 * ܂ReLXgԂŌĂ΂邽߁CVXeR[͗pĂ͂ȂD
 * (T[rXR[𗘗p邱)
 */
static int consdrv_intrproc(struct consreg *cons)
{
  unsigned char c;
  char *p;

  if (serial_is_recv_enable(cons->index)) { /* M */
    c = serial_recv_byte(cons->index);
    if (c == '\r') /* sR[hϊ(\r\n) */
      c = '\n';

    send_string(cons, &c, 1); /* GR[obN */

    if (cons->id) {
      if (c != '\n') {
	/* słȂȂCMobt@Ƀobt@O */
	cons->recv_buf[cons->recv_len++] = c;
      } else {
	/*
	 * EnterꂽCobt@̓e
	 * R}hXbhɒʒmD
	 * (݃nhȂ̂ŁCT[rXER[𗘗p)
	 */
	p = kx_kmalloc(CONS_BUFFER_SIZE);
	memcpy(p, cons->recv_buf, cons->recv_len);
	kx_send(MSGBOX_ID_CONSINPUT, cons->recv_len, p);
	cons->recv_len = 0;
      }
    }
  }

  if (serial_is_send_enable(cons->index)) { /* M */
    if (!cons->id || !cons->send_len) {
      /* Mf[^Ȃ΁CMI */
      serial_intr_send_disable(cons->index);
    } else {
      /* Mf[^Ȃ΁CM */
      send_char(cons);
    }
  }

  return 0;
}

/* ݃nh */
static void consdrv_intr(void)
{
  int i;
  struct consreg *cons;

  for (i = 0; i < CONSDRV_DEVICE_NUM; i++) {
    cons = &consreg[i];
    if (cons->id) {
      if (serial_is_send_enable(cons->index) ||
	  serial_is_recv_enable(cons->index))
	/* ݂Ȃ΁CݏĂяo */
	consdrv_intrproc(cons);
    }
  }
}

static int consdrv_init(void)
{
  memset(consreg, 0, sizeof(consreg));
  return 0;
}

/* Xbh̗v */
static int consdrv_command(struct consreg *cons, kz_thread_id_t id,
			   int index, int size, char *command)
{
  switch (command[0]) {
  case CONSDRV_CMD_USE: /* R\[EhCo̎gpJn */
    cons->id = id;
    cons->index = command[1] - '0';
    cons->send_buf = kz_kmalloc(CONS_BUFFER_SIZE);
    cons->recv_buf = kz_kmalloc(CONS_BUFFER_SIZE);
    cons->send_len = 0;
    cons->recv_len = 0;
    serial_init(cons->index);
    serial_intr_recv_enable(cons->index); /* MݗL(MJn) */
    break;

  case CONSDRV_CMD_WRITE: /* R\[ւ̕o */
    /*
     * send_string()ł͑Mobt@𑀍삵ĂēsȂ̂ŁC
     * r̂߂Ɋ݋֎~ɂČĂяoD
     */
    INTR_DISABLE;
    send_string(cons, command + 1, size - 1); /* ̑M */
    INTR_ENABLE;
    break;

  default:
    break;
  }

  return 0;
}

int consdrv_main(int argc, char *argv[])
{
  int size, index;
  kz_thread_id_t id;
  char *p;

  consdrv_init();
  kz_setintr(SOFTVEC_TYPE_SERINTR, consdrv_intr); /* ݃nhݒ */

  while (1) {
    id = kz_recv(MSGBOX_ID_CONSOUTPUT, &size, &p);
    index = p[0] - '0';
    consdrv_command(&consreg[index], id, index, size - 1, p + 1);
    kz_kmfree(p);
  }

  return 0;
}
